/*
 *  IPFWCHAINS-MIB group implementation - ipfwchains.c 
 *	This module reads the firewalling rules from the new (kernel 2.2.x)
 * 	ipchains firewalling code for Linux.
 *      It is also possible to zero the counters, flush chains,
 *      and reorder the rules in any chain to improve performance.
 *      Author : Didier.Dhaenens@rug.ac.be 
 *               Atlantis 
 *               University of Ghent 
 *               Belgium
 *  18/04/1999
 *  This code is placed under the GNU-GPL, see COPYING for details
 */

#include "../mib_module_config.h"
#include <config.h>
#include <sys/types.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
/* mibincl.h contains all the snmp specific headers to define the
   return types and various defines and structures. */

#include "../mibincl.h"
#include "ipfwchains.h"
#include "libipfwc.h"
#include "../util_funcs.h"


struct chop_rule /* Construct needed for CHainOPtimization */
{
	u_int64_t score;
	const struct ipfwc_fwrule *rule;
};


WriteMethod writeZero;
WriteMethod writeOptimize;                  
WriteMethod writeFlush;

static struct ipfwc_fwchain *chainnames;
static unsigned int num_chains;
static struct ipfwc_fwrule *rules;
static unsigned int num_rules;

struct icmp_names {
        const char *name;
        __u16 type;
        __u16 code_min, code_max;
        int use_code_min_for_testing;
};

const struct icmp_names icmp_codes[] = {
        { "echo-reply", 0, 0, 0xFFFF, TRUE },
        /* Alias */ { "pong", 0, 0, 0xFFFF, TRUE },

        { "destination-unreachable", 3, 0, 0xFFFF, FALSE },
        {   "network-unreachable", 3, 0, 0, TRUE },
        {   "host-unreachable", 3, 1, 1, TRUE },
        {   "protocol-unreachable", 3, 2, 2, TRUE },
        {   "port-unreachable", 3, 3, 3, TRUE },
        {   "fragmentation-needed", 3, 4, 4, TRUE },
        {   "source-route-failed", 3, 5, 5, TRUE },
        {   "network-unknown", 3, 6, 6, TRUE },
        {   "host-unknown", 3, 7, 7, TRUE },
        {   "network-prohibited", 3, 9, 9, TRUE },
        {   "host-prohibited", 3, 10, 10, TRUE },
        {   "TOS-network-unreachable", 3, 11, 11, TRUE },
        {   "TOS-host-unreachable", 3, 12, 12, TRUE },
        {   "communication-prohibited", 3, 13, 13, TRUE },
        {   "host-precedence-violation", 3, 14, 14, TRUE },
        {   "precedence-cutoff", 3, 15, 15, TRUE },

        { "source-quench", 4, 0, 0xFFFF, TRUE },

        { "redirect", 5, 0, 0xFFFF, FALSE },
        {   "network-redirect", 5, 0, 0, TRUE },
        {   "host-redirect", 5, 1, 1, TRUE },
        {   "TOS-network-redirect", 5, 2, 2, TRUE },
        {   "TOS-host-redirect", 5, 3, 3, TRUE },

        { "echo-request", 8, 0, 0xFFFF, TRUE },
        /* Alias */ { "ping", 8, 0, 0xFFFF, TRUE },

        { "router-advertisement", 9, 0, 0xFFFF, TRUE },

        { "router-solicitation", 10, 0, 0xFFFF, TRUE },

        { "time-exceeded", 11, 0, 0xFFFF, FALSE },
        /* Alias */ { "ttl-exceeded", 11, 0, 0xFFFF, FALSE },
        {   "ttl-zero-during-transit", 11, 0, 0, TRUE },
        {   "ttl-zero-during-reassembly", 11, 1, 1, TRUE },

        { "parameter-problem", 12, 0, 0xFFFF, FALSE },
        {   "ip-header-bad", 12, 0, 0, TRUE },
        {   "required-option-missing", 12, 1, 1, TRUE },

        { "timestamp-request", 13, 0, 0xFFFF, TRUE },

        { "timestamp-reply", 14, 0, 0xFFFF, TRUE },

        { "address-mask-request", 17, 0, 0xFFFF, TRUE },

        { "address-mask-reply", 18, 0, 0xFFFF, TRUE }
};   

/* Functions that do some useful number-to-name conversions */
static char *addr_to_host(struct in_addr *); 
static char *addr_to_dotted(struct in_addr *);
static char *addr_to_network(struct in_addr *); 
static char *addr_to_anyname(struct in_addr *);
static char *mask_to_dotted(struct in_addr *);
static char *port_to_service(int, unsigned short);
static char *service_to_string(__u16, __u16);   

static int rule_compare(const void *a, const void *b);
static int optimize(const struct ipfwc_fwrule *rulez, 
		     unsigned int num_rulez, 
		     const char* chainlabel);
static int score_chain(struct chop_rule *rulez, unsigned int num_rulez);
	/*********************
	 *
	 *  Kernel & interface information,
	 *   and internal forward declarations
	 *
	 *********************/

/* None needed */

	/*********************
	 *
	 *  Initialisation & common implementation functions
	 *
	 *********************/

struct variable4 ipfwchains_variables[] = {
    { IPFWCCHAININDEX,		ASN_INTEGER, 	RONLY, var_ipfwchains, 3, {1, 1, 1}},
    { IPFWCCHAINLABEL,       	ASN_OCTET_STR, 	RONLY, var_ipfwchains, 3, {1, 1, 2}},
    { IPFWCPOLICY,          	ASN_OCTET_STR,  RONLY, var_ipfwchains, 3, {1, 1, 3}},
    { IPFWCREFCNT,          	ASN_INTEGER,  	RONLY, var_ipfwchains, 3, {1, 1, 4}},
    { IPFWCPKTS,          	ASN_OCTET_STR,  RONLY, var_ipfwchains, 3, {1, 1, 5}},
    { IPFWCBYTES,          	ASN_OCTET_STR,  RONLY, var_ipfwchains, 3, {1, 1, 6}},
    { IPFWCZERO,		ASN_INTEGER,	RWRITE, var_ipfwchains, 3, {1, 1, 7}},
    { IPFWCFLUSH,		ASN_INTEGER,	RWRITE, var_ipfwchains, 3, {1, 1, 8}},
    { IPFWCOPTIMIZE, 		ASN_INTEGER,	RWRITE, var_ipfwchains, 3, {1, 1, 9}},
    { IPFWRRULEINDEX,          	ASN_INTEGER,  	RONLY, var_ipfwrules,  3, {2, 1, 1}},
    { IPFWRCHAIN,             	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 2}},
    { IPFWRPKTS,                ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 3}},
    { IPFWRBYTES,               ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 4}},
    { IPFWRTARGET,              ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 5}},
    { IPFWRPROT,               	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 6}},
    { IPFWRSOURCE,              ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 7}},
    { IPFWRDESTINATION,    	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 8}},
    { IPFWRPORTS,               ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 9}},
    { IPFWROPT,         	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 10}},
    { IPFWRIFNAME,              ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 11}},
    { IPFWRTOSA,               	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 12}},
    { IPFWRTOSX,               	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 13}},
    { IPFWRMARK,              	ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 14}},
    { IPFWROUTSIZE,             ASN_OCTET_STR,  RONLY, var_ipfwrules,  3, {2, 1, 15}}
};

oid ipfwchains_variables_oid[] = { 1,3,6,1,4,1,2021,13,3 };

void init_ipfwchains( void )
{
    REGISTER_MIB( "ipfwchains/ipfwchains", ipfwchains_variables, variable4,
			ipfwchains_variables_oid);

    DEBUGMSGTL(("ipfwchains", "Started ipchains-module...\n"));
    chainnames = ipfwc_get_chainnames(&num_chains);
    rules = ipfwc_get_rules(&num_rules,0);    
}

	/*********************
	 *
	 *  System specific implementation functions
	 *
	 *********************/

u_char * var_ipfwchains(
    struct variable *vp,
    oid     *name,
    size_t  *length,
    int     exact,
    size_t  *var_len,
    WriteMethod **write_method)
{
    static char string_value[256];
    __u64 cnt, cntkb, cntmb, cntgb;         

    /* When the element at position (1,1) of the table is requested, we refresh the table. 
     * Note that we don't have to free chainnames, because ipfwc_get_chainnames only does 
     * one malloc : when it is first called
     */    
    if (  (name[*length-1]==1) && (name[*length-2]==1)  )
    {
 DEBUGMSGTL(("ipfwchains", "ipfwchains: Initialising chaintable...\n"));
	chainnames = ipfwc_get_chainnames(&num_chains);
	if (chainnames==NULL) return NULL;
    }

    /* Use checkmib from util_funcs.c to check if the oid is correct */
    if (!checkmib(vp,name,length,exact,var_len,write_method,num_chains)){
        DEBUGMSGTL(("ipfwchains", "ipfwchains: Match failed...\n"));
   	return NULL;
    }

    switch (vp->magic){
	case IPFWCCHAININDEX:
	        long_return = name[*length-1];
	    	return (u_char *)&long_return;
	case IPFWCCHAINLABEL:
		*var_len = strlen(chainnames[name[*length-1]-1].label);
		return (u_char *) chainnames[name[*length-1]-1].label;
	case IPFWCPOLICY:
                *var_len = strlen(chainnames[name[*length-1]-1].policy);
                return (u_char *) chainnames[name[*length-1]-1].policy;  
	case IPFWCREFCNT:
		*var_len = sizeof(long int);
		long_return = chainnames[name[*length-1]-1].refcnt;
		return (u_char *)&long_return;	
	case IPFWCPKTS:
                cnt = chainnames[name[*length-1]-1].packets;
                        if (cnt > 99999) {
                                cntkb = (cnt + 500) / 1000;
                                if (cntkb > 9999) {
                                        cntmb = (cnt + 500000) / 1000000;
                                        if (cntmb > 9999) {
                                                cntgb = (cntmb + 500) / 1000;
                                                sprintf(string_value, "%lluG", cntgb);
                                        }
                                        else
                                                sprintf(string_value, "%lluM", cntmb);
                                } else
                                        sprintf(string_value, "%lluK", cntkb);
                        } else
                                sprintf(string_value, "%llu", cnt);   
		*var_len = strlen (string_value);
		return (u_char *)string_value; 
	case IPFWCBYTES:
                cnt = chainnames[name[*length-1]-1].bytes;
                        if (cnt > 99999) {
                                cntkb = (cnt + 500) / 1000;
                                if (cntkb > 9999) {
                                        cntmb = (cnt + 500000) / 1000000;
                                        if (cntmb > 9999) {
                                                cntgb = (cntmb + 500) / 1000;
                                                sprintf(string_value, "%lluG", cntgb);
                                        }
                                        else
                                                sprintf(string_value, "%lluM", cntmb);
                                } else
                                        sprintf(string_value, "%lluK", cntkb);
                        } else
                                sprintf(string_value, "%llu", cnt);    
                *var_len = strlen (string_value);
                return (u_char *)string_value;
	case IPFWCZERO: 
            	long_return = 0;
		*var_len = sizeof (long_return);
		*write_method = writeZero; 
            	return (u_char *)&long_return;      
	case IPFWCFLUSH:
                long_return = 0;
                *var_len = sizeof (long_return);
                *write_method = writeFlush;
                return (u_char *)&long_return;   
	case IPFWCOPTIMIZE:
                long_return = 0;
		*var_len = sizeof (long_return);
                *write_method = writeOptimize;
                return (u_char *)&long_return; 
	default:
	    DEBUGMSGTL(("ipfwchains", "unknown sub-id %d in var_ipfwchains\n", vp->magic));
	    break;
    }
    return NULL;
}

u_char * var_ipfwrules(
    struct variable *vp,
    oid     *name,
    size_t  *length,
    int     exact,
    size_t  *var_len,
    WriteMethod **write_method)
{
    static char string_value[256];
    static char buf[256];
    static struct protoent *protocol;
    unsigned short flags;
    __u64 cnt, cntkb, cntmb, cntgb;

    /* When the element at position (1,1) of the table is requested, we refresh the table. 
     * Note that we don't have to free rules, because ipfwc_get_rules only does 
     * one malloc : when it is first called
     */
    if (  (name[*length-1]==1)&&(name[*length-2]==1) )
    {
 DEBUGMSGTL(("ipfwchains", "ipfwchains: Initialising ruletable...\n"));
        rules = ipfwc_get_rules(&num_rules,0);
        if (rules==NULL) return NULL;
    }

    if (!checkmib(vp,name,length,exact,var_len,write_method,num_rules)){
        DEBUGMSGTL(("ipfwchains", "ipfwchains: Match failed...\n"));
        return NULL;
    }

    switch (vp->magic){
        case IPFWRRULEINDEX:
		long_return = name[*length-1];
		return (u_char *)&long_return;     
        case IPFWRCHAIN:
                *var_len = strlen(rules[name[*length-1]-1].chain[0].label);
                return (u_char *) rules[name[*length-1]-1].chain[0].label; 
        case IPFWRPKTS:
	 DEBUGMSGTL(("ipfwchains", "ipfwchains: case IPFWRPKTS\n"));
                cnt = rules[name[*length-1]-1].packets;
                        if (cnt > 99999) {
                                cntkb = (cnt + 500) / 1000;
                                if (cntkb > 9999) {
                                        cntmb = (cnt + 500000) / 1000000;
                                        if (cntmb > 9999) {
                                                cntgb = (cntmb + 500) / 1000;
                                                sprintf(string_value, "%lluG", cntgb);
                                        }
                                        else
                                                sprintf(string_value, "%lluM", cntmb);
                                } else
                                        sprintf(string_value, "%lluK", cntkb);
                        } else 
                                sprintf(string_value, "%llu", cnt);   
                *var_len = strlen (string_value);
                return (u_char *) string_value;   


        case IPFWRBYTES:
                cnt = rules[name[*length-1]-1].bytes;
                        if (cnt > 99999) {
                                cntkb = (cnt + 500) / 1000;
                                if (cntkb > 9999) {
                                        cntmb = (cnt + 500000) / 1000000;
                                        if (cntmb > 9999) {
                                                cntgb = (cntmb + 500) / 1000;
                                                sprintf(string_value, "%lluG", cntgb);
                                        }
                                        else
                                                sprintf(string_value, "%lluM", cntmb);
                                } else
                                        sprintf(string_value, "%lluK", cntkb);
                        } else  
                                sprintf(string_value, "%llu", cnt);  
                *var_len = strlen (string_value);
                return (u_char *)string_value;  
        case IPFWRTARGET:
                *var_len = strlen(rules[name[*length-1]-1].ipfw.label);
                return (u_char *) rules[name[*length-1]-1].ipfw.label;   
        case IPFWRPROT:
		protocol = getprotobynumber( (int) rules[name[*length-1]-1].ipfw.ipfw.fw_proto);
		if (!strcmp(protocol->p_name,"ip" )) strcpy (protocol->p_name,"all");
		*var_len = strlen (protocol->p_name);
		return (u_char *) protocol->p_name;
        case IPFWRSOURCE:
        	if (rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCIP)  sprintf(string_value,"!");
		else strcpy(string_value,"");   
        	if (rules[name[*length-1]-1].ipfw.ipfw.fw_smsk.s_addr == 0L)
                	strcat(string_value,"anywhere");
        	else {
                        sprintf(buf, "%s", addr_to_anyname(&(rules[name[*length-1]-1].ipfw.ipfw.fw_src)));
               	 	strcat(buf, mask_to_dotted(&(rules[name[*length-1]-1].ipfw.ipfw.fw_smsk)));
                	strcat(string_value,buf);
        		}
	 DEBUGMSGTL(("ipfwchains", "ipfwchains: %s\n",string_value));
		*var_len = strlen(string_value);
		return (u_char *) string_value;  
        case IPFWRDESTINATION:
                if (rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTIP)  sprintf(string_value,"!");
                else strcpy(string_value,"");
                if (rules[name[*length-1]-1].ipfw.ipfw.fw_dmsk.s_addr == 0L)
                        strcat(string_value,"anywhere");
                else {
                        sprintf(buf, "%s", addr_to_anyname(&(rules[name[*length-1]-1].ipfw.ipfw.fw_dst)));
                        strcat(buf, mask_to_dotted(&(rules[name[*length-1]-1].ipfw.ipfw.fw_dmsk)));
                        strcat(string_value,buf);
                        }
                DEBUGMSGTL(("ipfwchains", "ipfwchains: %s\n",string_value));
                *var_len = strlen(string_value);
                return (u_char *) string_value;      
        case IPFWRPORTS:
        	if (rules[name[*length-1]-1].ipfw.ipfw.fw_proto != IPPROTO_TCP
            		&& rules[name[*length-1]-1].ipfw.ipfw.fw_proto != IPPROTO_UDP
            		&& rules[name[*length-1]-1].ipfw.ipfw.fw_proto != IPPROTO_ICMP) {
                		sprintf(string_value,"n/a");
				*var_len = strlen(string_value);
                		return (u_char *) string_value;
        	}
 
        	/* ICMP handled specially. */
        	if (rules[name[*length-1]-1].ipfw.ipfw.fw_proto == IPPROTO_ICMP
            	   && !(rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCPT)
            	   && !(rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTPT)) {
                      unsigned int i;
                      for (i = 0; i < sizeof(icmp_codes)/sizeof(struct icmp_names); i++) {
                            if (icmp_codes[i].type == rules[name[*length-1]-1].ipfw.ipfw.fw_spts[0]
                               && icmp_codes[i].type == rules[name[*length-1]-1].ipfw.ipfw.fw_spts[1]
                               && icmp_codes[i].code_min == rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[0]
                               && icmp_codes[i].code_max == rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[1]) {
                                   sprintf(string_value, "%s", icmp_codes[i].name);
				   *var_len = strlen(string_value);  
				   return (u_char *) string_value;
                               }
                      }
        	}

        	sprintf(string_value, rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_SRCPT ? "!" : "");
        	if (rules[name[*length-1]-1].ipfw.ipfw.fw_spts[0] == 0 && rules[name[*length-1]-1].ipfw.ipfw.fw_spts[1] == 0xFFFF)
                   strcat(string_value, "any");
        	else if (rules[name[*length-1]-1].ipfw.ipfw.fw_spts[0] == rules[name[*length-1]-1].ipfw.ipfw.fw_spts[1]) {
		   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_spts[0], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));
        	}
        	else {
                   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_spts[0], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));   
		   strcat(string_value,":");
                   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_spts[1], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));  
        	}
		strcat (string_value," -> ");
                strcat(string_value, rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_DSTPT ? "!" : "");
                if (rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[0] == 0 && rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[1] == 0xFFFF)
                   strcat(string_value, "any");
                else if (rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[0] == rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[1]) {
                   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[0], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));
                }
                else {
                   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[0], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));
                   strcat(string_value,":");
                   strcat(string_value, service_to_string(rules[name[*length-1]-1].ipfw.ipfw.fw_dpts[1], rules[name[*length-1]-1].ipfw.ipfw.fw_proto));
                }    
                *var_len = strlen(string_value);
                return (u_char *) string_value;  
        case IPFWROPT:
		flags = rules[name[*length-1]-1].ipfw.ipfw.fw_flg;
                sprintf(string_value, (rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_SYN) ? "!" : "-");
                strcat(string_value, (flags & IP_FW_F_TCPSYN) ? "y" : "-");
                strcat(string_value, (rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_FRAG) ? "!" : "-");
                strcat(string_value, (flags & IP_FW_F_FRAG) ? "f" : "-");
                strcat(string_value, (flags & IP_FW_F_PRN) ? "l" : "-");
                strcat(string_value, (flags & IP_FW_F_NETLINK) ? "o" : "-");
                *var_len = strlen(string_value);
                return (u_char *) string_value;    
        case IPFWRIFNAME:
                sprintf(string_value, rules[name[*length-1]-1].ipfw.ipfw.fw_invflg & IP_FW_INV_VIA ? "!" : "");
                if (rules[name[*length-1]-1].ipfw.ipfw.fw_flg & IP_FW_F_WILDIF && (rules[name[*length-1]-1].ipfw.ipfw.fw_vianame)[0]) {
                        rules[name[*length-1]-1].ipfw.ipfw.fw_vianame[strlen(rules[name[*length-1]-1].ipfw.ipfw.fw_vianame)+1]='\0';
                        rules[name[*length-1]-1].ipfw.ipfw.fw_vianame[strlen(rules[name[*length-1]-1].ipfw.ipfw.fw_vianame)]='+';
                }
                strcat(string_value, (rules[name[*length-1]-1].ipfw.ipfw.fw_vianame)[0] ? rules[name[*length-1]-1].ipfw.ipfw.fw_vianame : "any");
                *var_len = strlen(string_value);
                return (u_char *) string_value;    
        case IPFWRTOSA:
                sprintf(string_value, "0x%02hX", (unsigned short) rules[name[*length-1]-1].ipfw.ipfw.fw_tosand);
                *var_len = strlen(string_value);
                return (u_char *) string_value;
        case IPFWRTOSX:
                sprintf(string_value, "0x%02hX", (unsigned short) rules[name[*length-1]-1].ipfw.ipfw.fw_tosxor);
                *var_len = strlen(string_value);
                return (u_char *) string_value;   
        case IPFWRMARK:
                if (rules[name[*length-1]-1].ipfw.ipfw.fw_flg & IP_FW_F_MARKABS)
                        sprintf(string_value, "0x%x",rules[name[*length-1]-1].ipfw.ipfw.fw_mark);
                else if (rules[name[*length-1]-1].ipfw.ipfw.fw_mark == 0)
                        strcpy(string_value,"");
                else
                        sprintf(string_value, "0x%x", (int)rules[name[*length-1]-1].ipfw.ipfw.fw_mark);
                *var_len = strlen(string_value);
                return (u_char *) string_value;     
        case IPFWROUTSIZE:
                if ((rules[name[*length-1]-1].ipfw.ipfw.fw_flg & IP_FW_F_NETLINK) && (rules[name[*length-1]-1].ipfw.ipfw.fw_outputsize != 0xFFFF))
                        sprintf(string_value, "%hu", rules[name[*length-1]-1].ipfw.ipfw.fw_outputsize);
                else
                        strcpy(string_value,"");
                *var_len = strlen(string_value);
                return (u_char *)string_value;
        default:
	    DEBUGMSGTL(("snmpd", "unknown sub-id %d in var_ipfwrules\n", vp->magic));
    }
    return NULL;
}             

int writeZero(int action,
             u_char *var_val,
             u_char var_val_type,
             size_t var_val_len,
             u_char *statP,
             oid *name,
             size_t name_len)
{
    int status;
    long tmp=0;
    int tmplen=1000;

    if (var_val_type != ASN_INTEGER){
        fprintf(stderr,"not integer\n");
        return SNMP_ERR_WRONGTYPE;
    }
    asn_parse_int(var_val,&tmplen,&var_val_type,&tmp,sizeof(int));  
    if ( (action == COMMIT) && ( tmp != 0 )  ){
	status = ipfwc_zero_entries(chainnames[name[name_len-1]-1].label); 
    }
    return SNMP_ERR_NOERROR;
} 


int writeFlush(int action,
             u_char *var_val,
             u_char var_val_type,
             size_t var_val_len,
             u_char *statP,
             oid *name,
             size_t name_len)
{
    int status;
    long tmp=0;
    int tmplen=1000;

    if (var_val_type != ASN_INTEGER) {
       printf("Wrong type != int\n");
       return SNMP_ERR_WRONGTYPE;
    }
    asn_parse_int(var_val,&tmplen,&var_val_type,&tmp,sizeof(int));
    if ( (action == COMMIT) && ( tmp != 0 )  ){   
        status = ipfwc_flush_entries(chainnames[name[name_len-1]-1].label);

    }
    return SNMP_ERR_NOERROR; 
}
      


int writeOptimize(int action,
             u_char *var_val,
             u_char var_val_type,
             size_t var_val_len,
             u_char *statP,
             oid *name,
             size_t name_len)
{
/* XX why not use the one copy ?
    unsigned int num_rules;
    struct ipfwc_fwrule *rules; 
*/
    long tmp=0;
    int tmplen=1000;          
    if (var_val_type != ASN_INTEGER){
        fprintf(stderr,"not integer\n");
        return SNMP_ERR_WRONGTYPE;
    }
    asn_parse_int(var_val,&tmplen,&var_val_type,&tmp,sizeof(int));     
    if ( (action == COMMIT) && ( tmp != 0  )  ){   
        if (rules==NULL) return 0; 
	rules = ipfwc_get_rules(&num_rules, 0);
        if (!rules) {
	    fprintf(stderr, "Could not read rules: %s", ipfwc_strerror(errno)); 
	    exit(1);
        }

	optimize(rules, num_rules, chainnames[name[name_len-1]-1].label);

    }
    return SNMP_ERR_NOERROR;   

	return 0;
}


	/*********************
	 *
	 *  Internal implementation functions
	 *
	 *********************/

static char* addr_to_host(struct in_addr *addr)
{
        struct hostent *host;
        if ((host = gethostbyaddr((char *) addr, sizeof(struct in_addr), AF_INET)) != NULL)
                return (char *) host->h_name;
        else
                return (char *) NULL;
}
 
static char* addr_to_network(struct in_addr *addr)
{
        struct netent *net;
        if ((net = getnetbyaddr((long) ntohl(addr->s_addr), AF_INET)) != NULL)
                return (char *) net->n_name;
        else
                return (char *) NULL;
}

static char* addr_to_anyname(struct in_addr *addr)
{
        char *name;
        if ((name = addr_to_host(addr)) != NULL)
                return name;
        else if ((name = addr_to_network(addr)) != NULL)
                return name;
        else
                return addr_to_dotted(addr);
}
 
static char* addr_to_dotted(struct in_addr *addrp)
{
        static char buf[20];
        unsigned char *bytep;
        bytep = (unsigned char *) &(addrp->s_addr);
        sprintf(buf, "%d.%d.%d.%d", bytep[0], bytep[1], bytep[2], bytep[3]);
        return buf;
}
 
static char* mask_to_dotted(struct in_addr *mask)
{
        int i;
        static char buf[20];
        __u32 maskaddr, bits;

        maskaddr = ntohl(mask->s_addr);
        if (maskaddr == 0xFFFFFFFFL) {
                /* we don't want to see "/32" */
                strcpy(buf, "");
                return buf;
        }
        else {
                i = 32;
                bits = 0xFFFFFFFEL;
                while (--i >= 0 && maskaddr != bits)
                        bits <<= 1;
                if (i >= 0)
                        sprintf(buf, "/%d", i);
                else
                        /* mask was not a decent combination of 1's and 0's */
                        sprintf(buf, "/%s", addr_to_dotted(mask));
                return buf;
        }
}
 
static char* port_to_service(int port, unsigned short proto)
{
        struct servent *service;

        if (proto == IPPROTO_TCP && (service = getservbyport(htons(port), "tcp")) != NULL)
                return service->s_name;
        else if (proto == IPPROTO_UDP &&
                        (service = getservbyport(htons(port), "udp")) != NULL)
                return service->s_name;
        else
                return (char *) NULL;
}
 
static char* service_to_string(__u16 port, __u16 proto)
{
	static char *service;
	static char buf[30];
	if ( (service=port_to_service(port,proto))==NULL){ 
		sprintf(buf,"%u",port);
		return buf;
	}
	return service;
}

static int optimize(const struct ipfwc_fwrule *rulez, 
		     unsigned int num_rulez, 
		     const char* chainlabel)
{
   struct chop_rule *rule_array;
   int rc = 0;
   unsigned int i, count;
   /* skip to specified chain */ 
   for (i = 0; 
	i < num_rulez && strcmp(chainlabel, rulez[i].chain->label) != 0; 
	i++);

   if (i == num_rulez) {
	   fprintf(stderr, "`%s' is empty or does not exist.\n", chainlabel);
	   return 0;
   }
   /* Make temporary chain... */
   if (!ipfwc_create_chain("__chop")) {
	   fprintf(stderr, "Error creating chain `__chop': %s\n",
		   ipfwc_strerror(errno));
	   return 0;
   }
   rule_array = malloc(sizeof(struct chop_rule)*num_rulez);

   /* Count number of rules, and append to temp chain. */
   for (count = 0;
	i+count < num_rulez 
		&& strcmp(chainlabel, rulez[i+count].chain->label) == 0;
	count++) {
	   rule_array[count].rule = &rulez[i+count];
	   DEBUGMSGTL(("ipfwchains","%s  ", rulez[i+count].ipfw.label));
	   if (!ipfwc_append_entry("__chop", &rulez[i+count].ipfw)) {
		   fprintf(stderr, "Error appending rule %d to chain `__chop': %s\n",
			   count,ipfwc_strerror(errno));
		   goto cleanup_tmp;
	   }
   }

   /* Get scores. */
   if (!score_chain(rule_array, count))
	   goto cleanup_tmp;

   /* Sort by score; best to worst. */
   qsort((void *)rule_array, count, sizeof(struct chop_rule), &rule_compare);

   /* Redirect packets to temporary chain. */
   {
	   struct ip_fwuser fw[2]
		   = { { { { 0 }, { 0 }, { 0 }, { 0 }, 0, 0, IP_FW_F_WILDIF, 0,
			   { 0, 0xFFFF }, { 0, 0xFFFF}, 0, 0, "", 0xFF, 0x00 },
			 "__chop" },
		       { { { 0 }, { 0 }, { 0 }, { 0 }, 0, 0, IP_FW_F_WILDIF, 0,
			   { 0, 0xFFFF }, { 0, 0xFFFF}, 0, 0, "", 0xFF, 0x00 },
			 IP_FW_LABEL_RETURN } };

	   if (!ipfwc_insert_entry(chainlabel, &fw[0], 1)) {
		   fprintf(stderr,
			   "Error inserting patch rule to " "__chop" ": %s\n",
			   ipfwc_strerror(errno));
		   goto cleanup_tmp;
	   }

	   if (!ipfwc_insert_entry(chainlabel, &fw[1], 2)) {
		   fprintf(stderr, "Error inserting RETURN rule: %s\n",
			   ipfwc_strerror(errno));
		   goto cleanup_patch;
	   }
   }

   /* Clean out existing rules. */
   while (ipfwc_delete_num_entry(chainlabel, 3));

   /* Append reordered rules. */
   for (i = 0; i < count; i++) {
	   if (!ipfwc_append_entry(chainlabel, &rule_array[i].rule->ipfw)) {
		   fprintf(stderr,
			   "Error inserting rule %u (target `%s'): %s\n",
			   i, rule_array[i].rule->ipfw.label,
			   ipfwc_strerror(errno));
		   /* Leave things as they are: not unsafe, just messy. */
		   goto cleanup_done;
	   }
   }

   /* Delete blocker rules. */
   ipfwc_delete_num_entry(chainlabel, 2);
   ipfwc_delete_num_entry(chainlabel, 1);

   rc = 1;
   goto cleanup_tmp;

cleanup_patch:
   ipfwc_delete_num_entry(chainlabel, 1);

cleanup_tmp:
   /* Delete temporary chain. */
   ipfwc_flush_entries("__chop");
   ipfwc_delete_chain("__chop");

cleanup_done:
   free(rule_array);
   return (rc);
}

static int rule_compare(const void *a, const void *b)
{
   const struct chop_rule *c, *d;
   c = (const struct chop_rule *)a;
   d = (const struct chop_rule *)b;
   if (c->score > d->score) return -1;
   if (c->score < d->score) return +1;   
   return 0;
}

#define BETWEEN(x,a,b) ((x) >= (a) && (x) <= (b))

static int overlap_single(u_int32_t a[2], u_int32_t b[2])
{
	/* If either is empty range, then they don't overlap. */
	return (a[1] >= a[0]
		&& b[1] >= b[0]
		&& (BETWEEN(a[0], b[0], b[1])
		    || BETWEEN(a[1], b[0], b[1])
		    || BETWEEN(b[0], a[0], a[1])));
}

static int overlap(u_int32_t a[2][2], u_int32_t b[2][2])
{
	return overlap_single(a[0], b[0])
		|| overlap_single(a[0], b[1])
		|| overlap_single(a[1], b[0])
		|| overlap_single(a[1], b[1]);
}

static void empty_range(u_int32_t range[2])
{
	range[0] = 1;
	range[1] = 0;
}

/* Get inclusive ranges. */
static void get_range(u_int32_t low, u_int32_t high, int invert, 
		      u_int32_t max, u_int32_t range[2][2])
{
	if (invert) {
		if (low == 0)
			empty_range(range[0]);
		else {
			range[0][0] = 0;
			range[0][1] = low - 1;
		}
		if (high == max)
			empty_range(range[1]);
		else {
			range[1][0] = high + 1;
			range[1][1] = max;
		}
	} else {
		range[0][0] = low;
		range[0][1] = high;

		empty_range(range[1]);
	}
}


/* Return TRUE if a depends on b.  b depends on a iff there is a
 * packet which can match both rules, and the rules have different
 * effects.  
 *
 * Safer to return TRUE than FALSE.
 */
static int depends(const struct ip_fwuser *a, const struct ip_fwuser *b)
{
	/* If they have the same effect, there's no dependency. */
	if (strcmp(a->label, b->label) == 0
	    && a->ipfw.fw_redirpt == b->ipfw.fw_redirpt
	    && a->ipfw.fw_outputsize == b->ipfw.fw_outputsize
	    && a->ipfw.fw_tosand == b->ipfw.fw_tosand
	    && a->ipfw.fw_tosxor == b->ipfw.fw_tosxor
	    && a->ipfw.fw_mark == b->ipfw.fw_mark
	    && (a->ipfw.fw_flg&(IP_FW_F_MARKABS|IP_FW_F_PRN|IP_FW_F_NETLINK)) 
	    == (b->ipfw.fw_flg&(IP_FW_F_MARKABS|IP_FW_F_PRN|IP_FW_F_NETLINK))){
		DEBUGMSGTL(("ipfwchains","Same effect `%s' vs `%s'.\n", a->label, b->label));
		return FALSE;
	}

	/* If they specify the same field, but with different values, they
	 * are also independent. */

	/* IP src/dst specifications.  Crap out on non-contiguous masks. */
	{
		u_int32_t range[2][2][2];
		u_int32_t mask;

		for (mask = 0xFFFFFFFF; mask; mask <<= 1)
			if (mask == ntohl(a->ipfw.fw_smsk.s_addr)) break;

		if (mask != ntohl(a->ipfw.fw_smsk.s_addr)) {
			fprintf(stderr, "Non-contiguous netmask %u.%u.%u.%u\n",
				ntohl(a->ipfw.fw_smsk.s_addr) >> 24,
				(ntohl(a->ipfw.fw_smsk.s_addr) >> 16) & 0xFF,
				(ntohl(a->ipfw.fw_smsk.s_addr) >> 8) & 0xFF,
				ntohl(a->ipfw.fw_smsk.s_addr) & 0xFF);
			exit(1);
		}

		for (mask = 0xFFFFFFFF; mask; mask <<= 1)
			if (mask == ntohl(b->ipfw.fw_smsk.s_addr)) break;

		if (mask != ntohl(b->ipfw.fw_smsk.s_addr)) {
			fprintf(stderr, "Non-contiguous netmask %u.%u.%u.%u\n",
				ntohl(b->ipfw.fw_smsk.s_addr) >> 24,
				(ntohl(b->ipfw.fw_smsk.s_addr) >> 16) & 0xFF,
				(ntohl(b->ipfw.fw_smsk.s_addr) >> 8) & 0xFF,
				ntohl(b->ipfw.fw_smsk.s_addr) & 0xFF);
			exit(1);
		}
		
/* added checks for non-contiguous destination-masks */

		for (mask = 0xFFFFFFFF; mask; mask <<= 1)
			if (mask == ntohl(a->ipfw.fw_dmsk.s_addr)) break;

		if (mask != ntohl(a->ipfw.fw_dmsk.s_addr)) {
			fprintf(stderr, "Non-contiguous netmask %u.%u.%u.%u\n",
				ntohl(a->ipfw.fw_dmsk.s_addr) >> 24,
				(ntohl(a->ipfw.fw_dmsk.s_addr) >> 16) & 0xFF,
				(ntohl(a->ipfw.fw_dmsk.s_addr) >> 8) & 0xFF,
				ntohl(a->ipfw.fw_dmsk.s_addr) & 0xFF);
			exit(1);
		}

		for (mask = 0xFFFFFFFF; mask; mask <<= 1)
			if (mask == ntohl(b->ipfw.fw_dmsk.s_addr)) break;

		if (mask != ntohl(b->ipfw.fw_dmsk.s_addr)) {
			fprintf(stderr, "Non-contiguous netmask %u.%u.%u.%u\n",
				ntohl(b->ipfw.fw_dmsk.s_addr) >> 24,
				(ntohl(b->ipfw.fw_dmsk.s_addr) >> 16) & 0xFF,
				(ntohl(b->ipfw.fw_dmsk.s_addr) >> 8) & 0xFF,
				ntohl(b->ipfw.fw_dmsk.s_addr) & 0xFF);
			exit(1);
		}

		get_range(ntohl(a->ipfw.fw_src.s_addr&a->ipfw.fw_smsk.s_addr),
			  ntohl(a->ipfw.fw_src.s_addr|~a->ipfw.fw_smsk.s_addr),
			  a->ipfw.fw_invflg & IP_FW_INV_SRCIP,
			  0xFFFFFFFF, range[0]);
			
		get_range(ntohl(b->ipfw.fw_src.s_addr&b->ipfw.fw_smsk.s_addr),
			  ntohl(b->ipfw.fw_src.s_addr|~b->ipfw.fw_smsk.s_addr),
			  b->ipfw.fw_invflg & IP_FW_INV_SRCIP,
			  0xFFFFFFFF, range[1]);

		if (!overlap(range[0], range[1])) {
			DEBUGMSGTL(("ipfwchains","Different source IP ranges.\n"));
			return FALSE;
		}

		get_range(ntohl(a->ipfw.fw_dst.s_addr&a->ipfw.fw_dmsk.s_addr),
			  ntohl(a->ipfw.fw_dst.s_addr|~a->ipfw.fw_dmsk.s_addr),
			  a->ipfw.fw_invflg & IP_FW_INV_DSTIP,
			  0xFFFFFFFF, range[0]);
			
		get_range(ntohl(b->ipfw.fw_dst.s_addr&b->ipfw.fw_dmsk.s_addr),
			  ntohl(b->ipfw.fw_dst.s_addr|~b->ipfw.fw_dmsk.s_addr),
			  b->ipfw.fw_invflg & IP_FW_INV_DSTIP,
			  0xFFFFFFFF, range[1]);

		if (!overlap(range[0], range[1])) {
			DEBUGMSGTL(("ipfwchains","Different destination IP ranges.\n"));
			return FALSE;
		}
	}

	/* If both protos non-zero... */
	if (a->ipfw.fw_proto && b->ipfw.fw_proto) {
		/* ... and same ... */
		if (a->ipfw.fw_proto == b->ipfw.fw_proto) {
			/* ... and one inverted, not dependent. */
			if ((a->ipfw.fw_invflg & IP_FW_INV_PROTO)
			    != (b->ipfw.fw_invflg & IP_FW_INV_PROTO)) {
				DEBUGMSGTL(("ipfwchains","Same protos, one inverted.\n"));
				return FALSE;
			}
		/* ... or different, and neither inverted... */
		} 
		else if (!(a->ipfw.fw_invflg & IP_FW_INV_PROTO)
			 && !(b->ipfw.fw_invflg & IP_FW_INV_PROTO)) {
			DEBUGMSGTL(("ipfwchains","Different protos.\n"));
			return FALSE;
		}
	}

	/* If one says TCP syn, and the other says ! TCP syn, they're
	 * independent. */
	if ((a->ipfw.fw_flg & IP_FW_F_TCPSYN) 
	    && (b->ipfw.fw_flg & IP_FW_F_TCPSYN)
	    && ((a->ipfw.fw_invflg & IP_FW_INV_SYN)
		!= (b->ipfw.fw_invflg & IP_FW_INV_SYN))) {
		DEBUGMSGTL(("ipfwchains","Different TCP SYN.\n"));
		return FALSE;
	}

	/* If one says FRAG, and the other says ! FRAG or specifies ports,
	   they're independent. */
	if ((a->ipfw.fw_flg & IP_FW_F_FRAG) && (b->ipfw.fw_flg & IP_FW_F_FRAG)
	    && ((a->ipfw.fw_invflg & IP_FW_INV_FRAG)
		!= (b->ipfw.fw_invflg & IP_FW_INV_FRAG))) {
		DEBUGMSGTL(("ipfwchains","Different FRAG.\n"));
		return FALSE;
	}

	if ((a->ipfw.fw_flg & IP_FW_F_FRAG) && !(b->ipfw.fw_flg & IP_FW_F_FRAG)
	    && (ntohs(b->ipfw.fw_spts[0]) != 0 
		|| ntohs(b->ipfw.fw_spts[1]) != 0xFFFF
		|| ntohs(b->ipfw.fw_dpts[0]) != 0
		|| ntohs(b->ipfw.fw_dpts[1]) != 0xFFFF)) {
		DEBUGMSGTL(("ipfwchains","A frag, B ports.\n"));
		return FALSE;
	}

	if ((b->ipfw.fw_flg & IP_FW_F_FRAG) && !(a->ipfw.fw_flg & IP_FW_F_FRAG)
	    && (ntohs(a->ipfw.fw_spts[0]) != 0 
		|| ntohs(a->ipfw.fw_spts[1]) != 0xFFFF
		|| ntohs(a->ipfw.fw_dpts[0]) != 0
		|| ntohs(a->ipfw.fw_dpts[1]) != 0xFFFF)) {
		DEBUGMSGTL(("ipfwchains","B frag, A ports.\n"));
		return FALSE;
	}

	/* If the specified ports don't overlap, independent. */
	{
		u_int32_t range[2][2][2];

		get_range(a->ipfw.fw_spts[0], a->ipfw.fw_spts[1],
			  a->ipfw.fw_invflg & IP_FW_INV_SRCPT,
			  65535, range[0]);

		get_range(b->ipfw.fw_spts[0], b->ipfw.fw_spts[1],
			  b->ipfw.fw_invflg & IP_FW_INV_SRCPT,
			  65535, range[1]);

		if (!overlap(range[0], range[1])) {
			DEBUGMSGTL(("ipfwchains","Source ports different.\n"));
			return FALSE;
		}

		get_range(a->ipfw.fw_dpts[0], a->ipfw.fw_dpts[1],
			  a->ipfw.fw_invflg & IP_FW_INV_DSTPT,
			  65535, range[0]);

		get_range(b->ipfw.fw_dpts[0], b->ipfw.fw_dpts[1],
			  b->ipfw.fw_invflg & IP_FW_INV_DSTPT,
			  65535, range[1]);

		if (!overlap(range[0], range[1])) {
			DEBUGMSGTL(("ipfwchains","Destination ports different.\n"));
			return FALSE;
		}
	}

/*
  Consider this table (X = independent, ie. a packet can match both)

        eth0  eth+ !eth0 !eth+  ppp0  ppp+ !ppp0 !ppp+  p    !p     p+   !p+

  eth0              2     6     1     5                 1           5    
  eth+                    3     5     4                 5           4     
  !eth0  2                                                                   
  !eth+  6     3                                                            
  ppp0   1     5                            2     6     1                 6
  ppp+   5     4                                  3     5                 7
  !ppp0                         2                                         
  !ppp+                         6     3                                   
  p      1                      1     5                      2            6
  !p                                                    2                  
  p+     5     4                                                          3
  !p+                           6     7                 6           3      

  So we return FALSE if:
  1) Neither are wildcards, both different, and neither inverted.
  2) Neither are wildcards, both identical, and one inverted.
  3) Both wildcards, both identical, one inverted.
  4) Both wildcards, neither substring of other, and neither inverted.
  5) One is a wildcard, not substring of the other, and neither inverted.
  6) One is a wildcard, is a substring of the other, and only it is inverted.
  7) Both are wildcards, one is a substring of the other, and only it inverted.
*/

	if (!(a->ipfw.fw_flg & IP_FW_F_WILDIF)
	    && !(b->ipfw.fw_flg & IP_FW_F_WILDIF)) {
		/* 1) Both different, neither inverted. */
		if (strcmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame) != 0) {
			if (!(a->ipfw.fw_invflg & IP_FW_INV_VIA) 
			    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
				return FALSE;
		} else {
			/* 2) Both identical, one inverted. */
			if ((a->ipfw.fw_invflg & IP_FW_INV_VIA) 
			    != (b->ipfw.fw_invflg & IP_FW_INV_VIA))
				return FALSE;
		}
	} 
	else if ((a->ipfw.fw_flg & IP_FW_F_WILDIF) 
		 && (b->ipfw.fw_flg & IP_FW_F_WILDIF)) {
		/* 3) Both identical, one inverted. */
		if (strcmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame) == 0
		    && ((a->ipfw.fw_invflg & IP_FW_INV_VIA) 
			!= (b->ipfw.fw_invflg & IP_FW_INV_VIA)))
			return FALSE;

		/* 4) Neither substring of other, neither inverted. */
		if (strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
			    strlen(a->ipfw.fw_vianame)) != 0
		    && strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
			       strlen(b->ipfw.fw_vianame)) != 0
		    && !(a->ipfw.fw_invflg & IP_FW_INV_VIA) 
		    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
			return FALSE;

		/* 7) One substring of other, it is inverted. */
		if (strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
			    strlen(a->ipfw.fw_vianame)) == 0
		    && (a->ipfw.fw_invflg & IP_FW_INV_VIA)
		    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
			return FALSE;
		if (strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
			    strlen(b->ipfw.fw_vianame)) == 0
		    && (b->ipfw.fw_invflg & IP_FW_INV_VIA)
		    && !(a->ipfw.fw_invflg & IP_FW_INV_VIA))
			return FALSE;
	} else {
		/* 5) wildcard not substring of the other, neither inverted. */
		if (a->ipfw.fw_flg & IP_FW_F_WILDIF) {
			if (strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
				    strlen(a->ipfw.fw_vianame)) != 0) {
				if (!(a->ipfw.fw_invflg & IP_FW_INV_VIA) 
				    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
					return FALSE;
			} else {
				/* 6) wildcard is a substring of the
				   other, and only it is inverted. */
				if ((a->ipfw.fw_invflg & IP_FW_INV_VIA) 
				    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
					return FALSE;
			}
		} else {
			if (strncmp(a->ipfw.fw_vianame, b->ipfw.fw_vianame,
				    strlen(b->ipfw.fw_vianame)) != 0) {
				if (!(a->ipfw.fw_invflg & IP_FW_INV_VIA) 
				    && !(b->ipfw.fw_invflg & IP_FW_INV_VIA))
					return FALSE;
			} else {
				/* 6) wildcard is a substring of the
				   other, and only it is inverted. */
				if (!(a->ipfw.fw_invflg & IP_FW_INV_VIA) 
				    && (b->ipfw.fw_invflg & IP_FW_INV_VIA))
					return FALSE;
			}
		}
	}

	return TRUE;
}

/* Score the chain we want.  This is order n**2, oh well. */
static int score_chain(struct chop_rule *rulez, unsigned int num_rulez)
{
	int i;

	/* Iterate backwards through rules. */
	for (i = num_rulez - 1; i >= 0; i--) {
		unsigned int j;

		/* Score = number of hits we get... */
		rulez[i].score = rulez[i].rule->packets;
		for (j = (unsigned int)i+1; j < num_rulez; j++) {
			/* ... plus number of hits of dependents + number
			   of dependents */
			if (depends(&rulez[j].rule->ipfw, 
				    &rulez[i].rule->ipfw)) {
				DEBUGMSGTL(("ipfwchains","Rule %i depends on rule %i.\n", j, i));
				rulez[i].score += rulez[j].score + 1;
				if (rulez[i].score < rulez[j].score + 1) {
					/* Overflow could make us screw up. */
					fprintf(stderr, 
						"chop: counter overflow.\n");
					return 0;
				}
			}
			else DEBUGMSGTL(("ipfwchains","Rule %i independent of %i.\n", j, i));
		}
		DEBUGMSGTL(("ipfwchains","Rule %i score: %llu.\n", i, rulez[i].score));
	}
	return 1;
}
